
/*********************************************************************
 This is an example for our nRF52 based Bluefruit LE modules

 Pick one up today in the adafruit shop!

 Adafruit invests time and resources providing this open source code,
 please support Adafruit and open-source hardware by purchasing
 products from Adafruit!

 MIT license, check LICENSE for more information
 All text above, and the splash screen below must be included in
 any redistribution
*********************************************************************/

/* This sketch show how to use BLEClientService and BLEClientCharacteristic
 * to implement a custom client that is used to talk with Gatt server on
 * peripheral.
 *
 * Note: you will need another feather52 running peripheral/custom_HRM sketch
 * to test with.
 */

/** This was modified to read a cadence/speed sensor instead of a heart rate monitor, also
    a real cadence and speed sensor (CSAC) is used so contrary to what is stated above we 
    dont need a second hrm52, but the text is left since it is required by the MIT licence
    */

#include <bluefruit.h>

/* HRM Service Definitions
 * Heart Rate Monitor Service:  0x180D
 * Heart Rate Measurement Char: 0x2A37 (Mandatory)
 * Body Sensor Location Char:   0x2A38 (Optional)
 */

/* CSAC Service Definitions
 * Cadence and speed Service        :  0x1816
 * Cadence and speed Characteristics: 0x2A5B (Mandatory)
 */

// Routines for pedelec on and off

// Starts with an "impossible" value
static int last_cadence=-1;
unsigned last_pedelec_time;
static int pedelec_status=0;

static const int pedelec_timeout=3000;
static const int redled=LED_BUILTIN;
static const int ada7=7;

static void pedelec_on()
{
  pedelec_status=1;
  last_pedelec_time=millis();
}

static void pedelec_off()
{
  pedelec_status=0;
}

static void pedelec_init()
{
  // Setup the local IO
  pinMode(redled, OUTPUT);
  pinMode(ada7, OUTPUT);

  // Off initially
  pedelec_off();
}

// This is where we "fool" the bike controller that we are an ordinary wired pedelec
// Our pedelec is 66% duty cycle on and we just pedal 5 magnets in 3x5x100 = 1500 ms
// this should be easy to adjust if your pedelec does something different,
// future fancy versions may also meassure the speed of the ble sensor and adjust
// this patterns speed if controllers all of a sudden turn intelligent :-)
static int pattern_ix=0;
static int pattern[]={ HIGH, HIGH, LOW, 255};
static void pedelec_generate_pattern()
{
  if (pedelec_status==0) return;
  if (pattern[pattern_ix] == 255)
  {
    pattern_ix=0;
  }
    
  digitalWrite(redled, pattern[pattern_ix]);

  // Cant get pin 17 directly on adafruit so we use 7 instead :-)
  digitalWrite(ada7, pattern[pattern_ix]);
  pattern_ix++;
}

static void pedelec_check_timeout()
{
  // We realistically assume that no-one is gonna spend 50 days on a bike until this overflows
  // however it may be naive to think the controller is gonna be reset more often than that...
  unsigned long t = millis();

  if (t > last_pedelec_time + pedelec_timeout)
  {
    pedelec_off();
  }

  /*
  Serial.print("timer: ");
  Serial.print(t);
  Serial.print(" ");
  Serial.println(last_pedelec_time);
  */
}

BLEClientService        csacs(UUID16_SVC_CYCLING_SPEED_AND_CADENCE);

BLEClientCharacteristic csacc(UUID16_CHR_CSC_MEASUREMENT);


void setup()
{
  Serial.begin(115200);
  while ( !Serial ) delay(10);   // for nrf52840 with native usb

  Serial.println("Bluefruit52 Central Custom csac Example");
  Serial.println("--------------------------------------\n");

  pedelec_init();

  // Initialize Bluefruit with maximum connections as Peripheral = 0, Central = 1
  // SRAM usage required by SoftDevice will increase dramatically with number of connections
  Bluefruit.begin(0, 1);

  Bluefruit.setName("Bluefruit52 Central");

  // Initialize csac client
  csacs.begin();

  // set up callback for receiving measurement
  csacc.setNotifyCallback(csac_notify_callback);
  csacc.begin();

  // Increase Blink rate to different from PrPh advertising mode
  Bluefruit.setConnLedInterval(250);

  // Callbacks for Central
  Bluefruit.Central.setDisconnectCallback(disconnect_callback);
  Bluefruit.Central.setConnectCallback(connect_callback);

  /* Start Central Scanning
   * - Enable auto scan if disconnected
   * - Interval = 100 ms, window = 80 ms
   * - Don't use active scan
   * - Filter only accept csac service
   * - Start(timeout) with timeout = 0 will scan forever (until connected)
   */
  Bluefruit.Scanner.setRxCallback(scan_callback);
  Bluefruit.Scanner.restartOnDisconnect(true);
  Bluefruit.Scanner.setInterval(160, 80); // in unit of 0.625 ms

  // Some bug with this, now we just take whatever comes in our way and (hopefully) disconnects
  // if it is not a csac
  //Bluefruit.Scanner.filterUuid(csacs.uuid);
  Bluefruit.Scanner.useActiveScan(false);
  Bluefruit.Scanner.start(0);                   // // 0 = Don't stop scanning after n seconds
}

void loop()
{
  pedelec_check_timeout();
  pedelec_generate_pattern();
  delay(100);
}

/**
 * Callback invoked when scanner pick up an advertising data
 * @param report Structural advertising data
 */
void scan_callback(ble_gap_evt_adv_report_t* report)
{
  // filterUuid does not seem to work with magene twin cadence & speed sensors :-(
  // Add some code here to enable mac address locking.
  int i;
  Serial.println("Address of device found: ");
  for (i=0;i<6;i++)
  {
    Serial.print(report->peer_addr.addr[i],HEX);
    Serial.print(" ");
  }
  Serial.println();
  
  // Since we configure the scanner with filterUuid()
  // Scan callback only invoked for device with csac service advertised
  // Connect to device with csac service in advertising
  Bluefruit.Central.connect(report);
}

/**
 * Callback invoked when an connection is established
 * @param conn_handle
 */
void connect_callback(uint16_t conn_handle)
{
  Serial.println("Connected");
  Serial.print("Discovering csac Service ... ");

  // If csac is not found, disconnect and return
  if ( !csacs.discover(conn_handle) )
  {
    Serial.println("Found NONE");

    // disconnect since we couldn't find csac service
    Bluefruit.disconnect(conn_handle);

    return;
  }

  // Once csac service is found, we continue to discover its characteristic
  Serial.println("Found it");
  
  Serial.print("Discovering Measurement characteristic ... ");
  if ( !csacc.discover() )
  {
    // Measurement chr is mandatory, if it is not found (valid), then disconnect
    Serial.println("not found !!!");  
    Serial.println("Measurement characteristic is mandatory but not found");
    Bluefruit.disconnect(conn_handle);
    return;
  }
  Serial.println("Found it");

  // Reaching here means we are ready to go, let's enable notification on measurement chr
  if ( csacc.enableNotify() )
  {
    Serial.println("Ready to receive csac Measurement value");
  }else
  {
    Serial.println("Couldn't enable notify for csac Measurement. Increase DEBUG LEVEL for troubleshooting");
  }
}

/**
 * Callback invoked when a connection is dropped
 * @param conn_handle
 * @param reason is a BLE_HCI_STATUS_CODE which can be found in ble_hci.h
 */
void disconnect_callback(uint16_t conn_handle, uint8_t reason)
{
  (void) conn_handle;
  (void) reason;

  Serial.print("Disconnected, reason = 0x"); Serial.println(reason, HEX);
}


/**
 * Hooked callback that triggered when a measurement value is sent from peripheral
 * @param chr   Pointer client characteristic that even occurred,
 *              in this example it should be csacc
 * @param data  Pointer to received data
 * @param len   Length of received data
 */
void csac_notify_callback(BLEClientCharacteristic* chr, uint8_t* data, uint16_t len)
{
  // If they are on cadence or speed mode does not matter so much, speed gives 7 bytes of data and the
  // cadence gives five. But you can always just check the second byte (with index one) for changes and
  // thus determine if the sensor is moving or not.
 
  Serial.print("csac Measurement: ");

  int i;
  for (i=0;i<len;i++)
  {
    Serial.print(data[i], HEX);
    Serial.print(" ");
  }
  Serial.println();

  if (last_cadence==-1)
  {
    // First time around
    last_cadence = data[1];
  }
  else
  {
    if (last_cadence != data[1])
    {
      // Have we moved? Activate pedelec
      pedelec_on();
      last_cadence = data[1];
    }
  }
}
